;*General comments:

;*Op sys (kernel) code begins at C000

;*Variables:

JUMP                            equ     $7E     ;Jump instruction
RTI_PV                          equ     $EB     ;RTI Pseudo Vector 
RTI_V                           equ     $FFF0   ;RTI Vector 
						
KERNEL                          equ     $C000   ;kernel code location
PCB_BASE                        equ     $C800   ;Address of first PCB
OVERFLOW_PCBS                   equ     $C830   ;Address of one past the
						;last PCB
KERNEL_PCB                      equ     $C7FA   ;PCB used to kick start
						;the first process

PCB_STATE                       equ     $0      ;The following are the 
PCB_ID                          equ     $1      ;offsets of each element
PCB_STACK_BASE                  equ     $2      ;in the PCB
PCB_STACK_POINTER               equ     $4
PCB_SLEEPER                     equ     $6

PCB_SIZE                        equ     7      ;size of each PCB

PROC_ABSENT                     equ     0      ;State for empty PCB 
PROC_READY                      equ     1      ;State for an active PCB 
PROC_SLEEP                      equ     2      ;State for a sleeping PCB
PROC_WAITING                    equ     3      ;State for a sleeping PCB


STACK_BASE                      equ     $E040  ;The first process stack
STACK_SIZE                      equ     $40    ;stack size for each process
NUM_PROCS                       equ     8      ;Maximum number of processes


;Q_DEFAULT                       equ     62     ;The default QUANTUM (2.03s)
Q_DEFAULT                       equ     1     
RTI_PRESCALER                   equ     %00000011       ;32ms rti 

PACTL                           equ     $1026   
TFLG2                           equ     $1025
TMSK2                           equ     $1024

;*************************************************************************

SEMA4                           equ     0
SEMA4_OWNER                     equ     1
SEMA4_QUEUE_BASE                equ     3
SEMA4_QUEUE_OVERFLOW            equ     5
SEMA4_FIRST_QUEUE               equ     7
SEMA4_LAST_QUEUE                equ     9
SEMA4_QUEUE_BASE_OFFSET         equ     11
SEMA4_QUEUE_OVERFLOW_OFFSET     equ     29

;*************************************************************************

		org     KERNEL                  
		lds     #$1ff                   ;This is the kernel stack
		sei
	
		ldaa    #Q_DEFAULT              ;Set up software counter.
		staa    QUANTUM                 ;RTI must occur "quantum"
		staa    SW_RTI_COUNTER          ;number of times before
						;it switches process.
		
		ldaa    #0                     
		staa    NEXT_PID                

		jsr     create_pcb              ;initializes the PCB's
		

;this code initializes the "init" process PCB. 
	
		ldy     #init
		jsr     start

;A fake PCB address was needed to "jumpstart" the scheduler.  The
;false address was placed before the first real PCB so when the first 
;context switch tries to save a stack, it will be saved in the fake PCB

		ldx     #KERNEL_PCB             
		stx     CURRENT_PCB             

		;ldaa    #JUMP           ;RTI pseudo vector setup
		;staa    RTI_PV
		;ldx     #kernel_interrupt
		;stx     RTI_PV+1
		
		
		ldx     #kernel_interrupt
		stx     RTI_V

		
		ldaa    #RTI_PRESCALER  ; Set up RTI
		staa    PACTL
		ldaa    #%01000000
		staa    TMSK2   
		staa    TFLG2
		cli
		bra     *


	
;*********************************************  
;* Function:    create_pcb   
;*
;* Purpose:     Initializes all PCBs  
;*
;* Author:      Matt Enwald     
;*              Paul Clarke     April 22, 1997
;*
;* Reviewed:    Kalle Anderson
;*              Jason Buttron   April 26, 1997 
;*
;* Input:       none
;* 
;*********************************************

create_pcb:     
		ldy     #NUM_PROCS                      ;Do for each process
		ldx     #PCB_BASE
		ldd     #STACK_BASE
loop:
	       
		clr     PCB_STATE,x                    ;Set state to 0 ABSENT
		clr     PCB_ID,x                       ;Set ID to 0
		std     PCB_STACK_POINTER,x            ;Set stack pointer (also done in start sr)
		std     PCB_STACK_BASE,x               ;Set stack base of each stack (remains constant)
		addd    #STACK_SIZE                    ;Go to next stack

		xgdx
		addd    #PCB_SIZE
		xgdx                                   ;Go to next PCB
		
		dey                                   ;Go to next proc #
		bne     loop
		
		rts             

;*********************************************  
;* Function:    kernel_interrupt
;*
;* Purpose:     Scheduler and dispatcher
;*
;* Author:      Kalle Anderson     
;*              Jason Buttron   April 22, 1997
;*
;* Reviewed:    Matt Enwald
;*              Paul Clarke     April 26, 1997 
;*
;* Input:       none
;* 
;* Global 
;* Variables:   CURRENT_PCB = memory addres of the PCB for
;*                            the active process.
;*              SW_RTI_COUNT = number of RTI left before context
;*                             switch.
;*
;*********************************************

kernel_interrupt:
		ldx     #PCB_BASE
sleep_loop:     
		ldaa    PCB_STATE,x
		cmpa    #PROC_SLEEP     ; check if process is sleeping
		bne     sl_next         
		dec     PCB_SLEEPER,x   ; decrement sleep counter
		bne     sl_next
		ldab    #PROC_READY     ; if sleep counter is zero,
		stab    PCB_STATE,x     ; wake up process

sl_next:        xgdx                    ; check next PCB for sleep
		addd    #PCB_SIZE
		xgdx
		cpx     #OVERFLOW_PCBS  ;check if last PCB
		blo     sleep_loop



		dec     SW_RTI_COUNTER  ;decrements the RTI counter
					;stored in memory                 
		bne     kernel_int_exit ;not time for context switch
			
		ldaa    QUANTUM         ;reinitialize the QUANTUM counter
		staa    SW_RTI_COUNTER  ;when context switch happens
		
		ldx     CURRENT_PCB
pcb_loop:                               ;finds the next process' PCB 
		xgdx
		addd    #PCB_SIZE
		xgdx
					
		cpx     CURRENT_PCB     ;if there were no processes found
		beq     kernel_int_exit ;exit the interrupt

		cpx     #OVERFLOW_PCBS  ;goes from last PCB to first
		blo     pcb_notlast
	 
		ldx     #PCB_BASE
pcb_notlast:                            
		ldaa    PCB_STATE,x

		cmpa    #PROC_READY     
		beq     pcb_found

		;cmpa    #PROC_ABSENT   ;checks PCB for a process
		;beq     pcb_loop
		;cmpa    #PROC_SLEEP
		;beq     pcb_loop
		;cmpa    #PROC_WAITING
		;beq     pcb_loop

		bra     pcb_loop
pcb_found:
		
;Once the next process' PCB is found, a context switch must occur.  This
;is handled by the dispatcher.

dispatcher:     
		ldy     CURRENT_PCB
		sts     PCB_STACK_POINTER,y  ;save current stack pointer
		
		lds     PCB_STACK_POINTER,x  ;load the new stack pointer
		stx     CURRENT_PCB

		
kernel_int_exit:
		ldaa    #%01000000      ;re-enables the RTI
		staa    TFLG2
		rti                     ;returns to new process


;*********************************************  
;* Function:    exit
;*
;* Purpose:     Called by a process when it is done
;*
;* Author:      Kalle Anderson     
;*              Jason Buttron   April 22, 1997
;*
;* Reviewed:    Matt Enwald
;*              Paul Clarke     April 26, 1997 
;*
;* Input:       none
;*
;*********************************************


exit:   
		ldx     CURRENT_PCB
		ldaa    #PROC_ABSENT
		staa    PCB_STATE,x
		bra     *

						
;*********************************************  
;* Function:    start
;*
;* Purpose:     Initializes the PCB for a process.
;*              This function will loop until it
;*              finds an open PCB.
;*
;* Author:      Kalle Anderson     
;*              Matt Enwald     April 26, 1997
;*
;* Reviewed:    Jason Buttron
;*              Paul Clarke     April 26, 1997 
;*
;* Input:       Y register contains the address 
;*              of the process
;* 
;*********************************************

start:  
		sei                     ;enter the critical section
		ldx     #PCB_BASE
start_loop:             
		ldaa    PCB_STATE,x     ;looks for an unused PCB
		cmpa    #PROC_ABSENT
		beq     found_open_pcb
	
		xgdx                    
		addd    #PCB_SIZE
		xgdx                

		cpx     #OVERFLOW_PCBS  ;if the current PCB is the last
		blo     start_loop      ;PCB, the next PCB becomes the
		ldx     #PCB_BASE       ;first PCB
		bra     start_loop
found_open_pcb: 
		
	
		ldaa    #PROC_READY
		staa    PCB_STATE,x     ;makes the current PCB occupied
	
		ldaa    NEXT_PID        ;stores PID number in PCB
		staa    PCB_ID,x        
		inca
		staa    NEXT_PID
		
		ldd     PCB_STACK_BASE,x ;creates a stack so that
					 ;when an RTI is called,
					 ;the correct information
					 ;will be available.
		xgdx                                    
	       

		dex             ;PC low
		sty     0,x     ;PC is set for start address of new process
		dex             ;PC high
				
		dex             ;Y register space
		dex             
		
		dex             ;X register space
		dex             
		
		dex             ;ACCA space
		dex             ;BCCB space
				
		clr     0,x     ;makes default CCR ($40)
		bset    0,x %01000000
		dex             ;CCR space                       
		
		xgdx            
		std     PCB_STACK_POINTER,x
		cli
		rts

;*********************************************  
;* Function:    sleep
;*
;* Purpose:     
;*              
;*              
;*
;* Author:      Kalle Anderson  April 27, 1997
;*
;* Reviewed:    Matt Enwald
;*              Jason Buttron  April 27, 1997 
;*
;* Input:       accumulator a = sleeptime
;* 
;*********************************************

sleep:  
	ldx     CURRENT_PCB     
	staa    PCB_SLEEPER,x   ; load sleeptime into PCB
	ldaa    #PROC_SLEEP     
	staa    PCB_STATE,x     ; set process's state to sleep
	rts


;***************************************************************
;wait
;sema4 is in accx
;CURRENT_PCB = my pcb
;***************************************************************

sema4_wait:
	sei     ;ATOMIC

	ldaa    SEMA4,x         ;value of sema4
	beq     grab_sema4

	inc     SEMA4,x         

	pshx
	ldy     SEMA4_LAST_QUEUE,x      ;insert process in sema4 
	ldx     CURRENT_PCB             ; circular queue
	stx     0,y
	iny                     ;increment last area in queue
	iny 
	pulx

	cpy     SEMA4_QUEUE_OVERFLOW,x  ;check if gone past end of 
	bne     wnot_overflow           ;circular queue
	ldy     SEMA4_QUEUE_BASE,x

wnot_overflow:
	sty     SEMA4_LAST_QUEUE,x

	ldx     CURRENT_PCB     ; set current process to wait state
	ldaa    #PROC_WAITING
	staa    PCB_STATE,x

	cli

waiter_loop:
	cmpa    PCB_STATE,x
	beq     waiter_loop     ;this just finishes off the quantum
	bra     wait_end

grab_sema4:
	inc     SEMA4,x         ;if sema4 not owned, gain ownership
	ldy     CURRENT_PCB
	sty     SEMA4_OWNER,x
	cli

wait_end:       
	rts


;***************************************************************
;release
;
;***************************************************************

sema4_release:
	sei
	ldy     CURRENT_PCB             ; check if current process
	cpy     SEMA4_OWNER,x           ; is the owner
	bne     release_end

	dec     SEMA4,x                 ; if sema4 becomes zero,
	beq     release_end             ; no process is waiting

	ldy     SEMA4_FIRST_QUEUE,x     ;load pointer to first pcb
	ldy     0,y                     ;load pcb into y 

	ldaa    #PROC_READY             ; make next in queue the
	staa    PCB_STATE,y             ; owner of the sema4
	sty     SEMA4_OWNER,x

	ldy     SEMA4_FIRST_QUEUE,x     ; increment next spot in
	iny                             ; circular queue
	iny
	cpy     SEMA4_QUEUE_OVERFLOW,x  ; check if overflow in 
	bne     rnot_overflow;          ; circular queue
	ldy     SEMA4_QUEUE_BASE,x
rnot_overflow:
	sty     SEMA4_FIRST_QUEUE,x
release_end:
	cli
	rts


;************
;sema4_init
;register x - holds semaphore address
;************
sema4_init:     
	ldaa    #0
	staa    0,x             ; Initialize sema4 to zero
	staa    SEMA4,x         
	pshx
	pula
	pulb

	addd    #SEMA4_QUEUE_BASE_OFFSET        ; setup circular
	std     SEMA4_FIRST_QUEUE,x             ; queue pointers
	std     SEMA4_LAST_QUEUE,x      
	std     SEMA4_QUEUE_BASE,x

	pshx
	pula
	pulb
	addd    #SEMA4_QUEUE_OVERFLOW_OFFSET    ; setup overflow values
	std     SEMA4_QUEUE_OVERFLOW,x
	rts

	
;kernel data     

QUANTUM         rmb     1       ;The RTI must occur "quantum" number 
				;times for the interupt to occur
SW_RTI_COUNTER  rmb     1       ;The number of RTIs left before the 
				;context switch
CURRENT_PCB     rmb     2       ;The address of active process' PCB
NEXT_PID        rmb     1       ;The PID number of the next process
STACK_POINTER   rmb     2       

LED_SEMA4       rmb     29
SEG7_SEMA4      rmb     29


;*********************************************  
;* Function:    init
;*
;* Purpose:     This process is started up automatically
;*              by the operating system. It starts up
;*              some other processes and then exits.
;*
;* Author:      Kalle Anderson  April 27, 1997
;*
;* Reviewed:    Matt Enwald
;*              Jason Buttron  April 27, 1997 
;*
;* Input:       none 
;* 
;*********************************************

init:
		ldx     #LED_SEMA4
		jsr     sema4_init

		ldy     #process_1a               
		jsr     start
		ldy     #process_1b               
		jsr     start
		ldy     #process_2               
		jsr     start
		jsr     exit
		

;*********************************************  
;* Function:    process_2
;*
;* Purpose:     Display a 4 digit counter in hex
;*
;* Author:      Dr. Mark Sebern
;*
;*********************************************

* LED display example
*
pioc    equ     $1002
portb   equ     $1004
ddrc    equ     $1007
portc   equ     $1003

	
process_2
	ldaa    pioc    Set display stobe (STRB) active low.
	anda    $fe
	staa    pioc
	ldaa    #3      Set C0 and C1 to outputs (display char select).
	staa    ddrc

	clr     num     Clear high order byte of number.
	clr     num+1   Clear low order byte of number.

nextnum
	ldd     num     Get number to temp area.
	std     num2
	ldaa    #0      Address character 0
	staa    portc
nextdig
	ldx     #hextbl Get address of hex table.
	ldab    num2+1  Get low byte of number.
	andb    #$0f    Mask low nibble.
	abx             Address correct hex character.
	ldaa    0,x     Get character.
	staa    portb
	ldd     num2    Get current number.
	lsrd            Shift to next nibble.
	lsrd
	lsrd
	lsrd
	std     num2    Save shifted number.
	inc     portc   Address next character.
	ldaa    portc   Test character address.
	anda    #$03
	bne     nextdig If not 00, do next digit.
	ldx     num     Increment number.
	inx
	stx     num

	ldd     num     
	andb    #$F
	bne     no_sleep_for_me

	ldaa    #$FF    
	jsr     sleep           ;I need a nap. :) zzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz

no_sleep_for_me:
	
	ldx     #0000   Waste time.
wait
	inx
	bne     wait

	bra     nextnum Go do next number.
hang        
	bra     hang    Wait for reset.

num     rmb     2       Number to display.
num2    rmb     2       Temp number storage.
hextbl  fcc     '0123456789ABCDEF'

;*********************************************  
;* Function:    process_1a
;*
;* Purpose:     To make LEDs flash
;*
;* Author:      Jason Buttron   April 26, 1997
;*
;* Reviewed:    Matt Enwald
;*              Kalle Anderson  April 26, 1997 
;*
;*Revised:      Jason Buttron   May 6,1997
;*
;* Input:       none 
;* 
;*********************************************

PORTA           equ     $1000
DELAY_TIME      equ     $FF

process_1a
p1loopaha  
	ldx     #LED_SEMA4
	jsr     sema4_wait

	ldaa    #$EF
	staa    PORTA
	ldx     #DELAY_TIME     ;light one LED
	jsr     delaya

	ldaa    #$DF            ;light a second LED
	staa    PORTA           ;and turn the previous LED off
	ldx     #DELAY_TIME
	jsr     delaya

	ldaa    #$BF
	staa    PORTA
	ldx     #DELAY_TIME
	jsr     delaya

	ldx     #LED_SEMA4
	jsr     sema4_release
	
	bra     p1loopaha

delaya  dex
	ldy     #$FFFF              
la      dey
	bne     la
	bne     delaya
	rts

;*********************************************  
;* Function:    process_1b
;*
;* Purpose:     To make LEDs flash
;*
;* Author:      Jason Buttron   May 6, 1997
;*
;* Reviewed:    Matt Enwald
;*              Kalle Anderson  May 6, 1997 
;*
;* Input:       none 
;* 
;*********************************************

process_1b
p1loopahb  
	ldx     #LED_SEMA4
	jsr     sema4_wait

	ldaa    #$BF
	staa    PORTA
	ldx     #DELAY_TIME     ;light one LED
	jsr     delayb

	ldaa    #$DF                    ;light a second LED
	staa    PORTA           ;and turn the previous LED off
	ldx     #DELAY_TIME
	jsr     delayb

	ldaa    #$EF
	staa    PORTA
	ldx     #DELAY_TIME
	jsr     delayb

	ldx     #LED_SEMA4
	jsr     sema4_release

	bra     p1loopahb

delayb  dex
	ldy     #$FFFF              
lb      dey
	bne     lb
	bne     delayb
	rts



