#include "hc11.h"
#include "hc11io.h"


////////////////////////////////////////////////////////////
// PortRegister
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: PortRegister
//
// Author: Paul Clarke
//
// Purpose: Constructor - Initializes the PortConnection member.
//
// Input Parameters: None.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
PortRegister::PortRegister()
{
	connect = NULL;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Read
//
// Author: Kalle Anderson et al.
//
// Purpose: Reads in the value of a port register.
//
// Input Parameters: None.
//
// Return Value: The byte value of the register.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
byte_t PortRegister::Read(void)
{
    Byte value;
    
    value.bit0(pins[0].GetLevel());
    value.bit1(pins[1].GetLevel());
    value.bit2(pins[2].GetLevel());
    value.bit3(pins[3].GetLevel());
    value.bit4(pins[4].GetLevel());
    value.bit5(pins[5].GetLevel());
    value.bit6(pins[6].GetLevel());
    value.bit7(pins[7].GetLevel());
    
    return value;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Read
//
// Author: Paul Clarke
//
// Purpose: To read in the level of a particular pin of a port.
//
// Input Parameters: PinNo - The pin number (0-7).
//
// Return Value: The level of the pin (0 or 1)
//
// Preconditions: PinNo must be between 0 and 7 inclusively.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
bit_t PortRegister::Read(int PinNo)
{
  assert((PinNo >=0) && (PinNo <= 7));
	return pins[PinNo].GetLevel();
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Kalle Anderson et al.
//
// Purpose: Writes a byte value to a port.
//
// Input Parameters: data - The byte to be written to the port.  Will update the
//                    PortConnection if one is attached and one of the pins is 
//                    an output pin.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortRegister::Write(byte_t data)
{
    Byte value;
    value = data;
	bit_t update = 0;

	  pins[0].Output(value.bit0());
    pins[1].Output(value.bit1());
    pins[2].Output(value.bit2());
    pins[3].Output(value.bit3());
    pins[4].Output(value.bit4());
    pins[5].Output(value.bit5());
    pins[6].Output(value.bit6());
    pins[7].Output(value.bit7());
/*	
	// If at least one of the pins is an output pin, the update the entire port
	for (int x = 0; x < 7; x++)
	{
		if (pins[x].GetMode() == Pin::OUTPUT)
		{
			update = 1;
			break;
		}
	}
*/
	if ((connect != NULL)) //&&(update)
		connect->Write(Read());
}
/*
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Input
//
// Author: Paul Clarke
//
// Purpose: To provide an input level to a particular pin of a port.  The level
//          of the pin will only be updated if the pin's direction is INPUT.
//          Will update the PortConnection if one is attached.
//
// Input Parameters: PinNo - The pin number (0-7).
//                    level - The new level of the pin.
//
// Return Value: None.
//
// Preconditions: PinNo must be between 0 and 7 inclusively.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortRegister::Input(int PinNo,bit_t level)
{
	assert((PinNo >=0) && (PinNo <= 7));
	// Pin::Input will check the direction of the pin
	pins[PinNo].Input(level);

	if (connect != NULL)
		connect->Write(Read());
}
*/
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Output
//
// Author: Paul Clarke
//
// Purpose: To provide an output level to a particular pin of a port. The level
//          of the pin will only change if the pin's direction mode is OUTPUT.
//          Will update the PortConnection if one is attached.
//
// Input Parameters: PinNo - The pin number (0-7)
//                    level - The new level of the pin.
//
// Return Value: None.
//
// Preconditions: PinNo must be between 0 and 7 inclusively.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortRegister::Output(int PinNo,bit_t level)
{
	assert((PinNo >=0) && (PinNo <= 7));
	// Pin::Input will check the direction of the pin
	pins[PinNo].Output(level);

	if (connect != NULL)
		connect->Write(Read());
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Action
//
// Author: Paul Clarke
//
// Purpose: To change the level of a pin of a port. The level of the pin will
//          only change if the pin's direction mode is AUTO.  Will update the
//          the PortConnection if one is attached.
//
// Input Parameters:  PinNo - The pin number (0-7).
//                    level - The new level of the pin.
//
// Return Value: None.
//
// Preconditions: PinNo must be betwee 0 and 7 inclusively.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortRegister::Action(int PinNo,bit_t level)
{
	assert((PinNo >=0) && (PinNo <= 7));
	// Pin::Input will check the direction of the pin
	pins[PinNo].Action(level);

	if (connect != NULL)
		connect->Write(Read());
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: PinInput
//
// Author: Paul Clarke
//
// Purpose: To provide an input level to a pin of a port. The level of the pin
//          will only change if the pin's direction mode is INPUT. Will update
//          the PortConnection if one is attached.
//
// Input Parameters:  PinNo - The pin number (0-7).
//                    level - The new level of the pin.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortRegister::PinInput(int PinNo, bit_t value)
{
	// Input just one pin
	assert((PinNo >= 0) && (PinNo <= 7));
 
	pins[PinNo].Input(value);

	if (connect != NULL)
		connect->Write(Read());
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Attach
//
// Author: Paul Clarke
//
// Purpose: To attach a PortConnection object to a port. If the input parameter
//          is NULL, it will not be attached.
//
// Input Parameters: pConnect - A pointer to a PortConnection object.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortRegister::Attach(PortConnection* pConnect)
{
	// Attach one output object to the entire port
	if (pConnect != NULL)
		connect = pConnect;
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Attach
//
// Author: Paul Clarke
//
// Purpose: To attach a PinConnection object to a pin on a port.
//
// Input Parameters:  PinNo - The pin number (0-7).
//                    pConnect - A pointer to a PinConnection object.
//
// Return Value: None.
//
// Preconditions: PinNo must be between 0 and 7 inclusively.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortRegister::Attach(int PinNo, PinConnection* pConnect)
{
	assert((PinNo >= 0) && (PinNo <= 7));
	
  if (pConnect != NULL)
	  pins[PinNo].Attach(pConnect);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: UnAttach
//
// Author: Kalle Anderson
//
// Purpose: To release the attachment of a PinConnection from a port's pin.
//
// Input Parameters:  PinNo - The pin number (0-7).
//
// Return Value: None.
//
// Preconditions: PinNo must be between 0 and 7 inclusively.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortRegister::UnAttach(int PinNo)
{
	assert((PinNo >= 0) && (PinNo <= 7));

	pins[PinNo].UnAttach();
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: SetMode
//
// Author: Paul Clarke
//
// Purpose: To set the direction mode of a port's pin.
//
// Input Parameters:  PinNo - The pin number (0-7).
//                    dir - The direction mode.
//
// Return Value: None.
//
// Preconditions: PinNo must be between 0 and 7 inclusively.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortRegister::SetMode(int PinNo, Pin::PinDirection dir)
{
	assert((PinNo >= 0) && (PinNo <= 7));

	pins[PinNo].SetMode(dir);
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: GetMode
//
// Author: Paul Clarke
//
// Purpose: To return the direciton mode of a port's pin.
//
// Input Parameters:  PinNo - The pin number (0-7)
//
// Return Value: None.
//
// Preconditions: PinNo must be between 0 and 7 inclusively.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
Pin::PinDirection PortRegister::GetMode(int PinNo)
{
	assert((PinNo >= 0) && (PinNo <= 7));

	return pins[PinNo].GetMode();
}

///////////////////////////////////////////////////////////////
// PortConnection
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Paul Clarke
//
// Purpose: To be used for the tesing of the PortConnection's functionality.
//          When used correctly, this function is overriden by a derived class
//          so this function should never be called.
//
// Input Parameters: data - The data to write to the PortConnection object.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortConnection::Write(byte_t data)
{

printf("Port Updated: %d %d %d %d %d %d %d %d\n",
					(data&0x80) ? 1 : 0,
					(data&0x40) ? 1 : 0,
					(data&0x20) ? 1 : 0,
					(data&0x10) ? 1 : 0,
					(data&0x8) ? 1 : 0,
					(data&0x4) ? 1 : 0,
					(data&0x2) ? 1 : 0,
					(data&0x1) ? 1 : 0);


}

///////////////////////////////////////////////////////////////////////////
// Pin Connection
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Set
//
// Author: Paul Clarke
//
// Purpose: Sets the level of a PinConnection.  When used correctly, this virtual
//          function is overriden by a derived class so this particular function
//          should never be called.
//
// Input Parameters: value - The level to be written to the PinConnection.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PinConnection::Set(bit_t value)
{
//	printf("Pin Updated %d\n",value);
}


///////////////////////////////////////////////////////////////
// DDR Register
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: DDRegister
//
// Author: Kalle Anderson
//
// Purpose: Initalize a DDRegister object.
//
// Input Parameters: None.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
DDRegister::DDRegister(void)
{
	port_reg = NULL;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Read
//
// Purpose: To read the byte value of a DDRegister.
//
// Input Parameters: None.
//
// Return Value: The byte value of the register.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
byte_t DDRegister::Read(void)
{
    return byte;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Paul Clarke
//
// Purpose: To write a byte value to the register.  After the byte is written,
//          the direction modes of the pins of the port that this register is
//          associated with are updated.
//
// Input Parameters: data - The byte to be written to the register.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void DDRegister::Write(byte_t data)
{
	byte = data;

	if (port_reg != NULL)
	{
		//Change the mode of the pins of the associated port
		if(port_reg->GetMode(0)!=Pin::AUTO)
			port_reg->pins[0].SetMode((data&0x1) ? Pin::OUTPUT : Pin::INPUT);
		if(port_reg->GetMode(1)!=Pin::AUTO)
			port_reg->pins[1].SetMode((data&0x2) ? Pin::OUTPUT : Pin::INPUT);
		if(port_reg->GetMode(2)!=Pin::AUTO)
			port_reg->pins[2].SetMode((data&0x4) ? Pin::OUTPUT : Pin::INPUT);
		if(port_reg->GetMode(3)!=Pin::AUTO)
			port_reg->pins[3].SetMode((data&0x8) ? Pin::OUTPUT : Pin::INPUT);
		if(port_reg->GetMode(4)!=Pin::AUTO)
			port_reg->pins[4].SetMode((data&0x10) ? Pin::OUTPUT : Pin::INPUT);
		if(port_reg->GetMode(5)!=Pin::AUTO)
			port_reg->pins[5].SetMode((data&0x20) ? Pin::OUTPUT : Pin::INPUT);
		if(port_reg->GetMode(6)!=Pin::AUTO)
			port_reg->pins[6].SetMode((data&0x40) ? Pin::OUTPUT : Pin::INPUT);
		if(port_reg->GetMode(7)!=Pin::AUTO)
			port_reg->pins[7].SetMode((data&0x80) ? Pin::OUTPUT : Pin::INPUT);
	}
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Associate
//
// Author: Paul Clarke
//
// Purpose: To associate a port with a DDRegister so that the DDRegister will
//          have control over the direction modes of the port's pins.
//
// Input Parameters:  port - Pointer to the port register to associate with the
//                    DDRegister.
//
// Return Value: None.
//
// Preconditions: The port pointer must not be NULL.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void DDRegister::Associate(PortRegister* port)
{
	assert(port != NULL);

	port_reg = port;
}



///////////////////////////////////////////////////////////////////
// Pin
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Input
//
// Author: Paul Clarke
//
// Purpose: To input a level to a pin.  The level of the pin will only change
//          if the pin's direction mode is INPUT. The PinConnection is updated
//          if one is attached.
//
// Input Parameters: value - The new level of the pin (0 or 1).
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void Pin::Input(bit_t value) 
{ 
	if (mode == Pin::INPUT) 
	{
		level = value;
		if (connect != NULL) 
			connect->Set(level); 
	}
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Output
//
// Author: Paul Clarke
//
// Purpose: To output a level to a pin.  The level of the pin will only change
//          if the pin's direction mode is OUTPUT.  The PinConnection will be
//          updated if one is attached.
//
// Input Parameters: value - The new level of the pin (0 or 1).
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void Pin::Output(bit_t value)	
{	
	if (mode == Pin::OUTPUT) 
	{
		level=value;
		if (connect != NULL) 
			connect->Set(level); 
	}
}
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Action
//
// Author: Paul Clarke
//
// Purpose: To set the level of a pin.  The level of the pin will only be 
//          changed if the pin's direction mode is AUTO.  The PinConnection will
//          be updated if one is attached.
//
// Input Parameters: None.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void Pin::Action(bit_t value)
{
	if (mode == AUTO)
	{
		level = value;
		if (connect != NULL)
			connect->Set(level);
	}
}

////////////////////////////////////////////////////////////////////////////
// Port "A" Register
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: PinInput
//
// Author: Kalle Anderson
//
// Purpose: To input a level to a pin of the Port A register and see if an
//          input compare or pulse accumulator occurred.  The level of the
//          pin will only be changed if the pin's direction mode is INPUT.
//          The PortConnection will be updated if one is attached.
//
// Input Parameters:  PinNo - The pin number (0-7).
//                    value - The new level of the pin.
//
// Return Value: None.
//
// Preconditions: PinNo must be between 0 and 7 inclusively.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void PortARegister::PinInput(int PinNo, bit_t value)
{
  // Input just one pin
	assert((PinNo >= 0) && (PinNo <= 7));

  
	bit_t previous_level = 0;
	// Edge mode determines what edges the IC's are sensitive to
	byte_t edge_mode = 0;

	// Edge type determines what edge (if any) has occurred
	byte_t edge_type = 0;

	
	// Edge configurations on EDGx
	//	EDGxA				EDGxB
	//		0						0					ICx's disabled
	//		0						1					Detect rising edges
	//		1						0					Detect falling edges
	//		1						1					Detect any edge


  switch(PinNo)
  {
      case 0:	// IC3
					// If the IC is enabled
          if (edge_mode = hc11->regfile.EDG3())
					{
						// If an edge occurred, a 0x3 does not represent an edge
						if ((edge_type = (Read(PinNo)<<1) | (value)) && (edge_type != 0x3))
						{
							// If the edge mode and type are equal, the right edge occurred (except 0x3)
							if ((edge_mode == edge_type) || (edge_mode == 0x3))
							{
								// Set flag
								hc11->regfile.IC3F(1);

								// Generate an interrupt (maybe)  ????

								// Set the TICx register
								hc11->regfile.TIC3.Write(hc11->regfile.TCNT.Read());
//printf("Input compare %d occurred at %d\n",3-PinNo,hc11->regfile.TCNT.Read());
//printf("TIC%d now holds %d\n",3-PinNo,hc11->regfile.TIC3.Read());
							}
						}		
					}

          pins[0].Input(value);
          break;

      case 1:	// IC2
					
					// If the IC is enabled
          if (edge_mode = hc11->regfile.EDG2())
					{
						// If an edge occurred, a 0x3 does not represent an edge
						if ((edge_type = (Read(PinNo)<<1) | (value)) && (edge_type != 0x3))
						{
							// If the edge mode and type are equal, the right edge occurred (except 0x3)
							if ((edge_mode == edge_type) || (edge_mode == 0x3))
							{
								// Set flag
								hc11->regfile.IC2F(1);

								// Generate an interrupt (maybe)  ????

								// Set the TICx register
								hc11->regfile.TIC2.Write(hc11->regfile.TCNT.Read());
//printf("Input compare %d occurred at %d\n",3-PinNo,hc11->regfile.TCNT.Read());
//printf("TIC%d now holds %d\n",3-PinNo,hc11->regfile.TIC2.Read());
							}
						}		
					}
          pins[1].Input(value);
          break;

      case 2:   // IC1
          
					// If the IC is enabled
          if (edge_mode = hc11->regfile.EDG1())
					{
						// If an edge occurred, a 0x3 does not represent an edge
						if ((edge_type = (Read(PinNo)<<1) | (value)) && (edge_type != 0x3))
						{
							// If the edge mode and type are equal, the right edge occurred (except 0x3)
							if ((edge_mode == edge_type) || (edge_mode == 0x3))
							{
								// Set flag
								hc11->regfile.IC1F(1);

								// Generate an interrupt (maybe)  ????

								// Set the TICx register
								hc11->regfile.TIC1.Write(hc11->regfile.TCNT.Read());
//printf("Input compare %d occurred at %d\n",3-PinNo,hc11->regfile.TCNT.Read());
//printf("TIC%d now holds %d\n",3-PinNo,hc11->regfile.TIC1.Read());
							}
						}		
					}  
				
					pins[2].Input(value);
          break;

      case 7:
          // PAI
          previous_level = Read(PinNo);            
          pins[PinNo].Input(value);

          //
          if((hc11->regfile.PAEN() && !hc11->regfile.PAMOD()) &&
             (previous_level != Read(PinNo)) &&
             (hc11->regfile.PEDGE() != previous_level))
          {                                   
              hc11->regfile.PAIF(1);                                 
              if(++hc11->regfile.PACNT == 0)
              {
                  hc11->regfile.PAOVF(1);  
              }
          }
          break;
  }
	

	if (connect != NULL)
		connect->Write(Read());
}


// Constructors for porta and portb

PortARegister::PortARegister()
{
	// Make Port A IC pins as inputs
	SetMode(0,Pin::INPUT);
	SetMode(1,Pin::INPUT);
	SetMode(2,Pin::INPUT);    
}

PortBRegister::PortBRegister()
{
  // Make Port B an output port
  for(int i=0;i<8;i++)
      SetMode(i, Pin::OUTPUT);   
}

////////////////////////////////////////////////////////////////////////////
// Port "C" Register
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Kalle Anderson
//
// Purpose: PortB needs some special stuff cause it is associated with
//          the handshake IO system.
//
// Input Parameters:  byte_t value
//                    
// Return Value: None.
/////////////////////////////////////////////////////////////////////////////

void PortBRegister::Write(byte_t data)
{
    PortRegister::Write(data);
    hc11->handshake_system->WriteToPORTB();
}
