///////////////////////////////////////////////////////////////
// File Name: hc11io.h
// Purpose:  Port/Pin object definitions
///////////////////////////////////////////////////////////////
#ifndef HC11IO_H
#define HC11IO_H

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include "basetypes.h"
#include "register.h"

//////////////////////////////////////////////////////////////////////////////
//
// Class Name: 
//
// Author: 
//
// Purpose: 
//
/////////////////////////////////////////////////////////////////////////////




//////////////////////////////////////////////////////////////////////////////
//
// Class Name: PinConnnection
//
// Author:  Paul Clarke et al.
//
// Purpose: Allows a pin to set a value in a different module (or thread)
//           by way of virtual functions.
//
/////////////////////////////////////////////////////////////////////////////
class PinConnection
{
public:
	// GUI writes this function
	virtual void Set(bit_t value);

};


//////////////////////////////////////////////////////////////////////////////
//
// Class Name: Pin
//
// Author: Paul Clarke, Kalle Anderson
//
// Purpose: This is the object that represents an outside pin of the HC11.
//          This is used by all the ports, by the XIRQ and the IRQ.
//          The pin can be attached to a PinConnection object.
//
/////////////////////////////////////////////////////////////////////////////
class Pin
{
 public:	
	typedef enum  { INPUT = 0, OUTPUT, AUTO } PinDirection;
 private:		
	bit_t level;
	PinDirection mode;
	PinConnection* connect;	
 public:	

    Pin() { connect = NULL; level = 0; mode = Pin::OUTPUT; };
	PinDirection GetMode() { return mode;};
	void SetMode(PinDirection value) { mode = value; };
	void Attach(PinConnection* pConnect) {	assert(pConnect != NULL); 
	                                        connect = pConnect; };
	void UnAttach(void) { connect = NULL;};

	// these should be changes later
	bit_t GetLevel(){return level;};
	
	void Input(bit_t value);	// Only changes pin if mode = INPUT

 protected:
	void Output(bit_t value); // Only changes pin if mode = OUTPUT
	void Action(bit_t value); // Only changes pin if mode = AUTO

	friend class PortRegister;
	friend class HC11SubSystem;
    friend class HandshakeIOSubSystem;
};


//////////////////////////////////////////////////////////////////////////////
//
// Class Name: PortConnection
//
// Author: Paul Clarke
//
// Purpose: Allows a port to write a value to an object in a different module
//          or thread via virtual functions.
//
/////////////////////////////////////////////////////////////////////////////
class PortConnection
{
public:	// Port writing to GUI
	// GUI writes this function
	virtual void Write(byte_t);

};

//////////////////////////////////////////////////////////////////////////////
//
// Class Name: PortRegister
//
// Author: Kalle Anderson
//
// Purpose: Handles all port operations and all of the operation of pins.
//
/////////////////////////////////////////////////////////////////////////////
class PortRegister: public MemoryRegister
{
 protected:
    Pin pins[8];
    PortConnection* connect;
 public:	

	PortRegister();

	// SIMULATOR INTERFACE TO PORT  
    byte_t Read(void);
	bit_t  Read(int PinNo);
	
	// These functions will update the pins
	void Write(byte_t data);

	// GUI INTERFACE TO PINS
 	void Attach(PortConnection* pConnect);
	void Attach(int PinNo, PinConnection* pConnect);
    void UnAttach(void){connect = NULL;};
    void UnAttach(int PinNo);

	virtual void PinInput(int PinNo,bit_t value);

	void SetMode(int PinNo, Pin::PinDirection dir);
	Pin::PinDirection GetMode(int PinNo);
    
    // !must make sure this can't be read/written to without using read/write!
    void Action(int PinNo,bit_t level);
	void Output(int PinNo,bit_t level);

	friend class DDRegister;
};

//////////////////////////////////////////////////////////////////////////////
//
// Class Name: PortARegister
//
// Author: Kalle Anderson
//
// Purpose: The override of a PortRegister so special events may happen
//          when Port A is changed.  This class is mainly used to capture
//          inputs on the pins.
//
/////////////////////////////////////////////////////////////////////////////
class PortARegister: public PortRegister
{
 private:
    class HC11 *hc11;
 public:	
    PortARegister();
    void SetHC11(HC11 *mcu){hc11 = mcu;};
    void PinInput(int PinNo,bit_t value);
	friend class DDRegister;
};



//////////////////////////////////////////////////////////////////////////////
//
// Class Name: PortARegister
//
// Author: Kalle Anderson
//
// Purpose: The override of a PortRegister so special events may happen
//          when Port A is changed.  This class is mainly used to capture
//          inputs on the pins.
//
/////////////////////////////////////////////////////////////////////////////
class PortBRegister: public PortRegister
{
 private:
    class HC11 *hc11;
 public:	
    PortBRegister();
    void SetHC11(HC11 *mcu){hc11 = mcu;};    
	void Write(byte_t data);
    friend class DDRegister;
};




//////////////////////////////////////////////////////////////////////////////
//
// Class Name: DDRegister
//
// Author: Kalle Anderson, Paul Clarke
//
// Purpose: These register classes affect the modes of pins of an 
//          associated port.
//
/////////////////////////////////////////////////////////////////////////////
class DDRegister: public ByteRegister
{
 private:
	 // This is the associated port register to this data
	 // direction register
	 PortRegister* port_reg; 
 public:	
	
	DDRegister();

	void Associate(PortRegister* port);
		
	byte_t Read(void);
	void Write(byte_t data);
};

#endif