//////////////////////////////////////////////////////////////////////////////
//  Filename:	regfile.cpp
//		
//    Purpose:	Contains the overriden functions for the register file.  All the
//    functions in this file are here because specific things happen
//    to the HC11 when bits of certain registers are set or cleared.
//
//    Author(s):	Paul Clarke, Kalle Anderson, Jason Buttron
//
//////////////////////////////////////////////////////////////////////////////

#include "regfile.h"
#include "memory.h"

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: <>
//
// Author: <>
//
// Purpose: <>
//
// Input Parameters: None.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////





///////////////////////////////////////////////////////////////
// RegisterFile functions

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: RegisterFile
//
// Author: Paul Clarke
//
// Purpose: Constructs a RegisterFile object and initializes it.
//
// Input Parameters: None.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
RegisterFile::RegisterFile()
{
	// Associate the DD registers to their ports
	DDRC.Associate(&PORTC);
	DDRD.Associate(&PORTD);

	// Set some pointer members to associated registers.
	TCTL1.portA = &PORTA;
	PACTL.portA = &PORTA;
	PACTL.oc1m = &OC1M;
	OC1M.portA = &PORTA;
	OC1M.tctl1 = &TCTL1;
	OC1M.pactl = &PACTL;
	CFORC.portA = &PORTA;
	CFORC.oc1m = &OC1M;
	CFORC.oc1d = &OC1D;
	CFORC.tctl1 = &TCTL1;
	SPCR.ddrd = &DDRD;
	SPCR.portD = &PORTD;

	// Make Port E an input port
  for(int i=0;i<8;i++)
      PORTE.SetMode(i, Pin::INPUT);   
  // port a and b configure themselves
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Reset
//
// Author: Paul Clarke
//
// Purpose: To set the registers to their reset state.
//
// Input Parameters: None.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void RegisterFile::Reset(Mode mode /*=SPECIAL_TEST*/)
{
	PORTA.Write(0x0);
	PIOC.Write(0x3);	// Bit 2 (PLS) is U
	PORTB.Write(0x0);	// Port B is an output port
	DDRC.Write(0x0);
	DDRD.Write(0x0);
	// Set the ports later, so the direction modes are initialized first
	PORTC.Write(0x0);
	PORTD.Write(0x0);
	PORTCL.Write(0x0);
	PORTE.Write(0x0);		// Port E is an input port
	CFORC.Write(0x0);
	OC1M.Write(0x0);
	OC1D.Write(0x0);
	TCNT.Write(0x0);
	// TICx registers not affected by reset
	TOC1.Write(0xffff);
	TOC2.Write(0xffff);
	TOC3.Write(0xffff);
	TOC4.Write(0xffff);
	TI4O5.Write(0xffff);
	TCTL1.Write(0x0);
	TCTL2.Write(0x0);	
	TMSK1.Write(0x0);
	TFLG1.Write(0x0);
	TMSK2.Write(0x0);
	TFLG2.Write(0x0);
	PACTL.Write(0x0);
	PACNT.Write(0x0);		// Completely U
	SPCR.Write(0x4);  // Bits 0,1 are U
	SPSR.Write(0x0);
	BAUD.Write(0x0);	// Bits 0,1,2 are U
	SCCR1.Write(0x0);	// Bits 6,7 are U
	SCCR2.Write(0x0);
	SCSR.Write(0xc0);
	SCDR.Write(0x0);	// Completely U
	ADCTL.Write(0x0); // Bits 0-5 are U
	ADR1.Write(0x0);	//  A/D input regs
	ADR2.Write(0x0);	
	ADR3.Write(0x0);	
	ADR4.Write(0x0);

	OPTION.Write(0x0);	// No delay
	COPRST.Write(0x0);	// COP register, don't care
	PPROG.Write(0x0);

	INIT.Write(0x1);
	
	// HPRIO reg on reset
	if (mode == SINGLE_CHIP)
		HPRIO.Write(0x5);
	else if (mode == EXPANDED)
		HPRIO.Write(0x25);
	else if (mode == BOOTSTRAP)
		HPRIO.Write(0xc5);
	else if (mode == SPECIAL_TEST)
		HPRIO.Write(0x75);

	TEST1.Write(0x0);
	CONFIG.Write(0xc);	// NOSEC and NOCOP high
}




/////////////////////////////////////////////////////////////////
// TCTL1 function overrides
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Paul Clarke
//
// Purpose: Writes to TCTL1 and sets modes of PortA pins as necessary
//
// Input Parameters: val - byte to be written to TCTL1 register.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void RegisterFile::TCTL1_C::Write(byte_t val)
{
	byte = val;

	// Check if any of the output compares are setup to change pins
	// If the output compares are enabled, set the pins's modes accordingly

	// OL2, OM2
	if (bit7() || bit6())
		portA->SetMode(6,Pin::AUTO);
	else
		portA->SetMode(6,Pin::OUTPUT);

	// OL3, OM3
	if (bit5() || bit4())
		portA->SetMode(5,Pin::AUTO);
	else 
		portA->SetMode(5,Pin::OUTPUT);

	// OL4, OM4
	if (bit3() || bit2())
		portA->SetMode(4,Pin::AUTO);
	else
		portA->SetMode(4,Pin::OUTPUT);

	// OL5, OM5
	if (bit1() || bit0())
		portA->SetMode(3,Pin::AUTO);
	else
		portA->SetMode(3,Pin::OUTPUT);
}

/////////////////////////////////////////////////////////////////
// PACTL function overrides
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Paul Clarke
//
// Purpose: Writes to the PACTL register and sets the mode of PA7 if necessary
//
// Input Parameters: val - byte to be written to the PACTL register.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void RegisterFile::PACTL_C::Write(byte_t val)
{
	byte = val;

	// Check DDRA7 to change the mode of PA7
	if (!bit7())
		portA->SetMode(7,Pin::INPUT);
	
	// If the pin is no longer an input pin, see if OC1 will now control it
	else if (oc1m->bit7())
		portA->SetMode(7,Pin::AUTO);
	else
		portA->SetMode(7,Pin::OUTPUT);
}

/////////////////////////////////////////////////////////////////
// OC1M function overrides
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Paul Clarke
//
// Purpose: Writes to the OC1M register and changes the modes of Port A pins
//					as necessary.
//
// Input Parameters: None.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void RegisterFile::OC1M_C::Write(byte_t val)
{
	byte = val;

	// The PA7 bit must be set to an output bit before the
	// pin can be changed to an automatic mode
	if (bit7())
	{
		if(portA->GetMode(7) == Pin::OUTPUT)
			portA->SetMode(7,Pin::AUTO);
	}
	else if (pactl->bit7())
		portA->SetMode(7,Pin::OUTPUT);
	else 
		portA->SetMode(7,Pin::INPUT);

	// Check to see if OC1 is controlling any other pins
	// If the individual output compares are not controlling the pin,
	// set it back to output
	if (bit6())
		portA->SetMode(6,Pin::AUTO);
	else if(!(tctl1->bit7() || tctl1->bit6()))
		portA->SetMode(6,Pin::OUTPUT);

	if (bit5())
		portA->SetMode(5,Pin::AUTO);
	else if(!(tctl1->bit5() || tctl1->bit4()))
		portA->SetMode(5,Pin::OUTPUT);

	if (bit4())
		portA->SetMode(4,Pin::AUTO);
	else if(!(tctl1->bit3() || tctl1->bit2()))
		portA->SetMode(4,Pin::OUTPUT);

	if (bit3())
		portA->SetMode(3,Pin::AUTO);
	else if(!(tctl1->bit1() || tctl1->bit0()))
		portA->SetMode(3,Pin::OUTPUT);

}

/////////////////////////////////////////////////////////////////
// INIT function overrides
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Paul Clarke, etc.
//
// Purpose: Writes to the INIT register and 
//
// Input Parameters: val - byte to write to the register
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void RegisterFile::INIT_C::Write(byte_t val)
{
	byte = val;
	
	byte_t temp = 0;

	// Check the movement of the register block
	if ((temp = byte<<4) != 0)
	{
		byte_t move_back = temp>>4;

		// The register block needs to be moved!
		RegisterMap* pregmap = (RegisterMap*)regmap;
		
		// Multiply the number by 4k to get the new starting address
		int new_base = move_back * 0x1000;

		// Change the memory object's addresses
		pregmap->Move(new_base);        
	}

	// Check the movement of the chip 256 byte RAM
	if ((temp = byte>>4) != 0)
	{
		// The on chip ram needs to be moved!
		RAM* pram = (RAM*)chip_ram;

		// Multiply the number by 4k to get the new starting address
		int new_base = temp * 0x1000;

		// Change the memory object's addresses
        pram->Move(new_base);
	}
}


void RegisterFile::HPRIO_C::Write(byte_t val)
{
    byte = val;
}


/////////////////////////////////////////////////////////////////
// TFLG1 and TFLG2
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Curious George
//
// Purpose: Writes to a TFLG register.  A bit is cleared in this register
//					if a one is written to it.
//
// Input Parameters: mask - The mask of what bits to clear
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void RegisterFile::TFLG_C::Write(byte_t mask)
{
	byte &= ~mask;
}

/*void RegisterFile::TFLG_C::Set(byte_t val)
{
byte = val;
}*/
	
/////////////////////////////////////////////////////////////////
// CFORC function overrides
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Paul Clarke
//
// Purpose: Writes to the CFORC register and possibly causes an output compare.
//
// Input Parameters: val - byte that is written to the CFORC register
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void RegisterFile::CFORC_C::Write(byte_t val)
{
	
	Byte b = val;

	// IMPORTANT 
	// CFORC can only be read as a zero, the bits never actually change 
	// levels.  A logical one written to a bit will fire an output compare.
	// A zero written to CFORC does nothing.
	if (byte != 0)
		byte = 0;

	// NOTE:  If OC1 activates a pin, other output compares cannot
	bit_t has_control2 = 0;
	bit_t has_control3 = 0;
	bit_t has_control4 = 0;
	bit_t has_control5 = 0;

	if (b.bit7())	// FOC1
	{
		// Check the control bits and set pin levels if necessary
		if (oc1m->bit3())
		{
			portA->Action(3,oc1d->bit3());
			has_control5 = 1;
		}
		if (oc1m->bit4())
		{
			portA->Action(4,oc1d->bit4());
			has_control4 = 1;
		}
		if (oc1m->bit5())
		{
			portA->Action(5,oc1d->bit5());
			has_control3 = 1;
		}
		if (oc1m->bit6())
		{
			portA->Action(6,oc1d->bit6());
			has_control2 = 1;
		}
		if (oc1m->bit7())
			portA->Action(7,oc1d->bit7());
	}

	if (b.bit6() && !has_control2)	// FOC2
	{
		bit_t ol2 = tctl1->bit6();
		bit_t om2 = tctl1->bit7();
	
		// Check the output compare enabling bits
		switch((om2<<1)|ol2)
		{
			case 0x0:		// does not affect pin
				break;
			case 0x1:		// toggles pin
				portA->Action(6,!portA->Read(6));
				break;
			case 0x2:		// clears pin
				portA->Action(6,0);
				break;
			case 0x3:		// sets pin
				portA->Action(6,1);
				break;
			default:	assert(false);
		}	
	}

	if (b.bit5() && !has_control3)	// FOC3
	{
		bit_t ol3 = tctl1->bit4();
		bit_t om3 = tctl1->bit5();
	
		// Check the output compare enabling bits
		switch((om3<<1)|ol3)
		{
			case 0x0:		// does not affect pin
				break;
			case 0x1:		// toggles pin
				portA->Action(5,!portA->Read(5));
				break;
			case 0x2:		// clears pin
				portA->Action(5,0);
				break;
			case 0x3:		// sets pin
				portA->Action(5,1);
				break;
			default:	assert(false);
		}
	}

	if (b.bit4() && !has_control4)	// FOC4
	{
		bit_t ol4 = tctl1->bit2();
		bit_t om4 = tctl1->bit3();
	
		// Check the output compare enabling bits
		switch((om4<<1)|ol4)
		{
			case 0x0:		// does not affect pin
				break;
			case 0x1:		// toggles pin
				portA->Action(4,!portA->Read(4));
				break;
			case 0x2:		// clears pin
				portA->Action(4,0);
				break;
			case 0x3:		// sets pin
				portA->Action(4,1);
				break;
			default:	assert(false);
		}
	}

	if (b.bit3() && !has_control5)	// FOC5
	{
		bit_t ol5 = tctl1->bit0();
		bit_t om5 = tctl1->bit1();
	
		// Check the output compare enabling bits
		switch((om5<<1)|ol5)
		{
			case 0x0:		// does not affect pin
				break;
			case 0x1:		// toggles pin
				portA->Action(3,!portA->Read(3));
				break;
			case 0x2:		// clears pin
				portA->Action(3,0);
				break;
			case 0x3:		// sets pin
				portA->Action(3,1);
				break;
			default:	assert(false);
		}
	}
}

/////////////////////////////////////////////////////////////////
// SPCR function overrides
//////////////////////////////////////////////////////////////////////////////
//
// Function Name: Write
//
// Author: Paul Clarke, Jason Buttron
//
// Purpose: Writes a byte to the SPCR register, updates the mode of Port D
//					pins as necessary.
//
// Input Parameters: val - byte to be written to the SPCR register.
//
// Return Value: None.
//
// Preconditions: None.
//
// Post Conditions: None.
//
/////////////////////////////////////////////////////////////////////////////
void RegisterFile::SPCR_C::Write(byte_t val)
{
	byte = val;

	// Check the DWOM bit to see if PORTD is being controlled by
  // the SCI

  // THIS WAS COMMENTED OUT BECAUSE THE SCI ONLY MAKES THE PORTD
  // PINS AS OPEN_DRAIN PINS, NOT PINS THAT CANNOT BE WRITTEN TO

/*	if (bit5())
	{
		// Take control of the PORTD pins
		for (int i = 0; i < 6; i++)
		{
			portD->SetMode(i,Pin::AUTO);
		}
	}
	else 
	{
		// Disable the AUTO setting we setup before
		for (int x = 0; x < 6; x++)
		{
			portD->SetMode(x,Pin::OUTPUT);
		}

		// Now go by the DDRD
		ddrd->Write(ddrd->Read());
	}
  */
}