///////////////////////////////////////////////////////////////
// File Name: register.h
// Purpose:  Register Definition
///////////////////////////////////////////////////////////////

#ifndef REGISTER_H
#define REGISTER_H

#include "basetypes.h"

//////////////////////////////////////////////////////////////////////////////
//
// Class Name: MemoryRegister
//
// Author: Kalle Anderson et al.
//
// Purpose: This class is a common register that every memory-mapped register
//					is derived from.  This allows the memory map to read and write to
//					all registers without knowing the specific type of register.
//
/////////////////////////////////////////////////////////////////////////////
class MemoryRegister
{    
 public:
	 virtual byte_t Read(void)=0;
	 virtual void Write(byte_t data)=0;	
};

//////////////////////////////////////////////////////////////////////////////
//
// Class Name: ByteRegister
//
// Author: Kalle Anderson et al.
//
// Purpose: A Byte object that has Read and Write operations.
//
/////////////////////////////////////////////////////////////////////////////
class ByteRegister: public Byte, public MemoryRegister
{    
 public:	
    byte_t Read(void){return byte;};
    void Write(byte_t data){byte = data;};

    ByteRegister &operator=(byte_t data){byte = data;return *this;};  

    byte_t operator++(int){return(++byte);}
    byte_t operator++(){return(++byte);}
};

//////////////////////////////////////////////////////////////////////////////
//
// Class Name: WordRegister
//
// Author: Kalle Anderson et al.
//
// Purpose: A 16-bit register from two ByteRegisters combined that has 
//					Read and Write operations.
//
/////////////////////////////////////////////////////////////////////////////
class WordRegister
{
 public:		
    // all this is is a container for two ByteRegisters
	ByteRegister	low;        
	ByteRegister	high;

	word_t Read(void){return ((high<<8)|(low));};
    void Write(word_t data){low = data & 0xFF; high = (data & 0xFF00)>>8;};

    word_t operator++(int) //Word++
    {                                
        word_t val = ((((word_t)high)<<8)|low) + 1;
        high = (val>>8);
        low = (val&0xFF);
        return val;
    }
    word_t operator++()   //++Word
    {                                
        word_t val = ((((word_t)high)<<8)|low) + 1;
        high = (val>>8);
        low = (val&0xFF);
        return val;
    }
};

#endif
